function output = starmap_solver(par)
%STARMAP_SOLVER
%   A second order staggered grid finite difference solver for
%   linear hyperbolic moment approximations (such as P_N and
%   SP_N) to the equations of radiative transfer in a 2D
%   geometry. Needs to be called from an example file by
%   STARMAP_SOLVER(PAR), where PAR defines the problem
%   parameters and system matrices.
%   OUTPUT = STARMAP_SOLVER(PAR) writes the final state of the
%   zeroth moment and its (x,y) coordinates into the struct
%   OUTPUT.
%
%   Version 1.1
%   Copyright (c) 07/14/2013 Benjamin Seibold and Martin Frank
%   http://www.math.temple.edu/~seibold
%   http://www.mathcces.rwth-aachen.de/5people/frank/start

%========================================================================
% License
%========================================================================
%   Permission is hereby granted, free of charge, to any person obtaining
%   a copy of this software and associated documentation files (the
%   "Software"), to deal in the Software without restriction, including
%   without limitation the rights to use, copy, modify, merge, publish,
%   distribute, sublicense, and/or sell copies of the Software, and to
%   permit persons to whom the Software is furnished to do so, subject to
%   the following conditions:
%   * The above copyright notice and this permission notice shall be
%     included in all copies or substantial portions of the Software.
%   * Appropriate references shall be given to the name of the Software
%     (StaRMAP), the website of the Software
%     (http://www.math.temple.edu/~seibold/research/starmap)
%     and the research article (http://arxiv.org/abs/1211.2205)
%     corresponding to the Software.
%
%   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
%   EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
%   MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
%   NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
%   BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
%   ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
%   CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
%   SOFTWARE.

%========================================================================
% Parameter Defaults
%========================================================================
if ~isfield(par,'Mx')||~isfield(par,'My')                  % If no system
    par.Mx = [0 1 0;1 0 0;0 0 0]/sqrt(3); par.n_mom = 1;       % matrices
    par.My = par.Mx([1 3 2],[1 3 2]);                % specified, use P1.
end
n_sys = length(par.Mx);                    % Number of system components.
if ~isfield(par,'name'),     par.name = 'Output'; end
if ~isfield(par,'ic'),       par.ic = @zero; end
if ~isfield(par,'sigma_a'),  par.sigma_a = @zero; end
if ~isfield(par,'sigma_s0'), par.sigma_s0 = @zero; end
flag_ani_scatter = isfield(par,'sigma_sm')&&isfield(par,'mom_order');
if ~flag_ani_scatter           % If anisotropic part of scattering kernel
    par.mom_order = [0 ones(1,n_sys-1)]; par.sigma_sm = @zero;      % not
end                    % available, set up data for isotropic scattering.
if ~isfield(par,'source'),  par.source = @zero; end
if ~isfield(par,'source_ind') % If no source moment vector specified, use
    par.source_ind = 1:((nargin(par.source)>=0)+...   % zeroth moment for
        (nargin(par.source)>3)*(n_sys-1));   % isotropic, and all moments
end                                             % for anisotropic source.
if ~isfield(par,'ax'),       par.ax = [0 1 0 1]; end
if ~isfield(par,'n'),        par.n = [100 100]; end
if ~isfield(par,'cn'),       par.cn = 0.99; end     % Default CFL number.
if ~isfield(par,'bc'),       par.bc = [0 0]; end
if ~isfield(par,'t_plot'),   par.t_plot = 0:.1:1; end
if ~isfield(par,'tfinal'),   par.tfinal = par.t_plot(end); end
if ~isfield(par,'output'),   par.output = @default_output; end
if ~isfield(par,'mom_output'), par.mom_output = 1; end % Moments plotted.
if strcmp(par.mom_output,'all'), par.mom_output = 1:n_sys; end

%========================================================================
% Initialization
%========================================================================
% Create index matrices and vectors for grid type
Ix = cellfun(@find,num2cell(par.Mx',1),'Un',0);         % Grid dependency
Iy = cellfun(@find,num2cell(par.My',1),'Un',0);                % indices.
c11 = 1; c22 = []; c21 = []; c12 = [];% First component is on (1,1)-grid.
while length([c11,c22,c21,c12])<n_sys                         % Determine
    c21 = unique(vertcat(c21',Ix{c11},Iy{c22})');            % indices of
    c12 = unique(vertcat(c12',Iy{c11},Ix{c22})');            % grid types
    c11 = unique(vertcat(c11',Ix{c21},Iy{c12})');           % from system
    c22 = unique(vertcat(c22',Iy{c21},Ix{c12})');             % matrices.
end
if isempty(c22), c22 = []; end
gtx = ones(1,n_sys); gty = gtx;                              % Grid types
gtx([c21,c22]) = 2; gty([c12,c22]) = 2;                     % in x and y.

% Grid size and time step
h = (par.ax([2 4])-par.ax([1 3]))./par.n;                    % Grid size.
dt = min(h)/abs(eigs(par.Mx,1,'lm'))/2*par.cn;       % Maximal time step.
dt = par.tfinal/ceil(par.tfinal/dt);    % Ensure integer number of steps.

% Create staggered grids and coefficients
x{1} = par.ax(1)+h(1)/2:h(1):par.ax(2)-h(1)/2;              % Cell center
y{1} = par.ax(3)+h(2)/2:h(2):par.ax(4)-h(2)/2;             % coordinates.
x{2} = par.ax(1)+h(1)*(1-par.bc(1)):h(1):par.ax(2);           % Staggered
y{2} = par.ax(3)+h(2)*(1-par.bc(2)):h(2):par.ax(4);        % coordinates.
[X,Y] = cellfun(@ndgrid,[x;x]',[y;y],'Un',0);          % Staggered grids.
n1 = par.n; n2 = size(X{2,2});                % Sizes of staggered grids.
extendx = {[1:par.bc(1),1:n1(1),par.bc(1)*(n1(1)-1)+1],...    % Extension
    [n2(1)*(1:1-par.bc(1)),1:n2(1)]};                       % indices for
extendy = {[1:par.bc(2),1:n1(2),par.bc(2)*(n1(2)-1)+1],...     % boundary
    [n2(2)*(1:1-par.bc(2)),1:n2(2)]};                       % conditions.
sg = [gtx;gty;par.mom_order]';                 % Construct staggered grid 
sg = unique(sg(2:end,:),'rows')';            % indices with moment order.

% Initialize field variables
sS = cell(2,2,max(par.mom_order)); sT = sS; ET = sS; % Scattering fields.
U = cell(1,n_sys); dxU = U; dyU = U; Q = U;   % Unknowns and derivatives.
for j = 1:n_sys                           % If initial condition only two 
    U{j} = X{gtx(j),gty(j)}*0+...                        % arguments, set
        capargs(par.ic,X{gtx(j),gty(j)},Y{gtx(j),gty(j)},j)*...     % all
        (nargin(par.ic)>2||j==1);               % higher moments to zero.
    Q{j} = X{gtx(j),gty(j)}*0;      % Initialize source terms with zeros.
end

%========================================================================
% Time Loop
%========================================================================
par.t_plot = [par.t_plot inf]; plot_count = 1;
cputime = zeros(1,3);
for t = dt:dt:par.tfinal
    U0 = U(par.mom_output);                     % Store current solution.
    for j = par.source_ind     % Source: evaluate only active components.
        Q{j} = capargs(par.source,X{gtx(j),gty(j)},...  % Evaluate source
            Y{gtx(j),gty(j)},t-dt/2,j);           % at half-time of step.
    end
    % Evaluate material parameters
    tic
    evaluate = (t==dt|[nargin(par.sigma_a)>2,nargin(par.sigma_s0)>2,...
        nargin(par.sigma_sm)>3])&[1 1 flag_ani_scatter];
    if evaluate(1)                                % Absorption: evaluate, 
        sA = cellfun(@(x,y)capargs(...                % if time-dependent
            par.sigma_a,x,y,t-dt/2),X,Y,'Un',0);    % or first time step.
    end
    if evaluate(2)                     % Scattering, isotropic component:
        for j = 1:size(sg,2)       % evaluate, if time-dependent or first
            sS{sg(1,j),sg(2,j),sg(3,j)} = capargs(par.sigma_s0,... % time
                X{sg(1,j),sg(2,j)},Y{sg(1,j),sg(2,j)},t-dt/2);    % step.
        end
    end
    if evaluate(3)                  % Scattering, anisotropic components:
        for j = 1:size(sg,2)             % evaluate, if time-dependent or
            sS{sg(1,j),sg(2,j),sg(3,j)} = ...          % first time step.
                sS{sg(1,j),sg(2,j),sg(3,j)}-capargs(par.sigma_sm,...
                X{sg(1,j),sg(2,j)},Y{sg(1,j),sg(2,j)},sg(3,j),t-dt/2);
        end
    end
    if any(evaluate)     % Total sigma_t^m = sigma_a+sigma_s^0-sigma_s^m.
        for j = 1:size(sg,2)                 % (Re)compute, if any of the 
            sT{sg(1,j),sg(2,j),sg(3,j)} = ...      % above (re)evaluated.
                sA{sg(1,j),sg(2,j)}+sS{sg(1,j),sg(2,j),sg(3,j)};
            ET{sg(1,j),sg(2,j),sg(3,j)} = ...        % Total sigma decay.
                expm1div(-sT{sg(1,j),sg(2,j),sg(3,j)}*dt/2);
        end
        EA = expm1div(-sA{1,1}*dt/2);     % Decay term for zeroth moment.
    end
    cputime(2) = cputime(2)+toc;
    % Update solution
    for step = [1 2 1]                   % Strang splitting of sub-steps.
        switch step
        case 1, tic               % Half step with (2,1) and (1,2) grids.
        for j = c11       % Compute update at (2,1) and (1,2) from (1,1).
            dxU{j} = diff(U{j}(extendx{1},:),[],1)/h(1);
            dyU{j} = diff(U{j}(:,extendy{1}),[],2)/h(2);
        end
        for j = c22       % Compute update at (1,2) and (2,1) from (2,2).
            dxU{j} = diff(U{j}(extendx{2},:),[],1)/h(1);
            dyU{j} = diff(U{j}(:,extendy{2}),[],2)/h(2);
        end
        for j = [c21 c12]   % Update components on (2,1) and (1,2) grids.
            W = -sumcell([dxU(Ix{j}),dyU(Iy{j})],...
                [par.Mx(j,Ix{j}),par.My(j,Iy{j})]);
            U{j} = U{j}+dt/2*(W+Q{j}-...
                sT{gtx(j),gty(j),par.mom_order(j)}.*U{j}).*...
                ET{gtx(j),gty(j),par.mom_order(j)};
        end
        cputime(1) = cputime(1)+toc;
        case 2, tic          % Two half steps with (1,1) and (2,2) grids.
        for j = c21       % Compute update at (1,1) and (2,2) from (2,1).
            dxU{j} = diff(U{j}(extendx{2},:),[],1)/h(1);
            dyU{j} = diff(U{j}(:,extendy{1}),[],2)/h(2);
        end
        for j = c12       % Compute update at (2,2) and (1,1) from (2,1).
            dxU{j} = diff(U{j}(extendx{1},:),[],1)/h(1);
            dyU{j} = diff(U{j}(:,extendy{2}),[],2)/h(2);
        end
        for j = [c11 c22]   % Update components on (1,1) and (2,2) grids.
            W = -sumcell([dxU(Ix{j}),dyU(Iy{j})],...
                [par.Mx(j,Ix{j}),par.My(j,Iy{j})]);
            for k = 1:2                         % Perform two half-steps.
                if j==1        % Zeroth moment decays by absorption only.
                    U{j} = U{j}+dt/2*(W+Q{j}-sA{1,1}.*U{j}).*EA;
                else                  % All other moments decay normally.
                    U{j} = U{j}+dt/2*(W+Q{j}-...
                        sT{gtx(j),gty(j),par.mom_order(j)}.*U{j}).*...
                        ET{gtx(j),gty(j),par.mom_order(j)};
                end
            end
        end
        cputime(1) = cputime(1)+toc;
        end
    end
    % Plotting
    tic
    while t>=par.t_plot(plot_count)-1e-14  % If current time has exceeded
        lambda = (par.t_plot(plot_count)-t+dt)/dt; %plotting time, define
        Uplot = cellfun(@(x,y)(1-lambda)*x+lambda*y,...    % solution via
            U0,U(par.mom_output),'Un',0); % linear interpolation in time.
        xplot = x(gtx(par.mom_output));                 % Assign grids at
        yplot = y(gty(par.mom_output));              % outputted moments.
        if length(par.mom_output)==1   % If only a single moment plotted,
            xplot = xplot{:}; yplot = yplot{:}; Uplot = Uplot{:};% remove
        end                                             % cell structure.
        if nargout(par.output)                              % Call output
            par = par.output(par,xplot,yplot,Uplot,plot_count);% routine,
        else                                  % allowing for it to modify
            par.output(par,xplot,yplot,Uplot,plot_count)% the struct par.
        end
        plot_count = plot_count+1;
    end
    cputime(3) = cputime(3)+toc;
end
fprintf('%s with %s%d\n %d moments, %dx%d grid, %0.0f time steps\n',... 
    par.name,par.closure,par.n_mom,n_sys,par.n,par.tfinal/dt)% Print test
cputime = reshape([cputime;cputime/sum(cputime)*1e2],1,[]);   % case info
fprintf(['CPU-times\n advection:%12.2fs%5.0f%%\n ',...   % and CPU times.
    'absorb/scatter:%7.2fs%5.0f%%\n plotting:%13.2fs%5.0f%%\n'],cputime)

% Output final state of solution
if nargout, output = struct('x',x(gtx),'y',y(gty),'U',U); end

%========================================================================
% Technical Functions
%========================================================================
function S = sumcell(A,w)
% Add vector of cells A, weighted with vector w.
S = A{1}*w(1); for j = 2:length(w), S = S+A{j}*w(j); end

function z = capargs(fct,varargin)
% Call function fct with as many arguments as it requires (at least 1),
% and ignore further arguments.
narg = max(nargin(fct),1);
z = fct(varargin{1:narg});

function f = zero(varargin)
% Zero function.
f = zeros(size(varargin{1}));

function y = expm1div(x)
% Function (exp(x)-1)/x that is accurate for x close to zero.
y = 1+x*.5+x.^2/6;
ind = abs(x)>2e-4;
y(ind) = (exp(x(ind))-1)./x(ind);

function default_output(par,x,y,U,step)
% Default plotting routine.
clf, imagesc(x,y,U'), axis xy equal tight
title(sprintf('%s at t = %0.2f',par.name,par.t_plot(step))), drawnow
